@@ -1,4 +1,4 @@
-{-# LANGUAGE FlexibleInstances, TypeSynonymInstances, TypeOperators, ScopedTypeVariables #-}
+{-# LANGUAGE FlexibleInstances, TypeSynonymInstances, TypeOperators, ScopedTypeVariables, NamedFieldPuns #-}
 
 -- | This module provides functions for calling command line programs, primarily
 --   'command' and 'cmd'. As a simple example:
@@ -10,24 +10,24 @@
 --   The functions from this module are now available directly from "Development.Shake".
 --   You should only need to import this module if you are using the 'cmd' function in the 'IO' monad.
 module Development.Shake.Command(
-    command, command_, cmd, unit, CmdArguments,
+    command, command_, cmd, unit, CmdArguments, (:->),
     Stdout(..), Stderr(..), Stdouterr(..), Exit(..), Process(..), CmdTime(..), CmdLine(..),
     CmdResult, CmdString, CmdOption(..),
     addPath, addEnv,
     ) where
 
 import Data.Tuple.Extra
-import Control.Applicative
 import Control.Exception.Extra
 import Control.Monad.Extra
 import Control.Monad.IO.Class
+import Data.Char
 import Data.Either.Extra
 import Data.List.Extra
 import Data.Maybe
 import System.Directory
 import System.Environment.Extra
 import System.Exit
-import System.IO.Extra
+import System.IO.Extra hiding (withTempFile, withTempDir)
 import System.Process
 import System.Info.Extra
 import System.Time.Extra
@@ -35,37 +35,20 @@ import System.IO.Unsafe(unsafeInterleaveIO)
 import qualified Data.ByteString as BS
 import qualified Data.ByteString.Lazy.Char8 as LBS
 import General.Process
+import Control.Applicative
 import Prelude
 
+import Development.Shake.CmdOption
 import Development.Shake.Core
 import Development.Shake.FilePath
+import Development.Shake.FilePattern
 import Development.Shake.Types
 import Development.Shake.Rules.File
+import Development.Shake.Derived(writeFile', withTempDir)
 
 ---------------------------------------------------------------------
 -- ACTUAL EXECUTION
 
--- | Options passed to 'command' or 'cmd' to control how processes are executed.
-data CmdOption
-    = Cwd FilePath -- ^ Change the current directory in the spawned process. By default uses this processes current directory.
-    | Env [(String,String)] -- ^ Change the environment variables in the spawned process. By default uses this processes environment.
-    | AddEnv String String -- ^ Add an environment variable in the child process.
-    | AddPath [String] [String] -- ^ Add some items to the prefix and suffix of the @$PATH@ variable.
-    | Stdin String -- ^ Given as the @stdin@ of the spawned process. By default the @stdin@ is inherited.
-    | StdinBS LBS.ByteString -- ^ Given as the @stdin@ of the spawned process.
-    | Shell -- ^ Pass the command to the shell without escaping - any arguments will be joined with spaces. By default arguments are escaped properly.
-    | BinaryPipes -- ^ Treat the @stdin@\/@stdout@\/@stderr@ messages as binary. By default 'String' results use text encoding and 'ByteString' results use binary encoding.
-    | Traced String -- ^ Name to use with 'traced', or @\"\"@ for no tracing. By default traces using the name of the executable.
-    | Timeout Double -- ^ Abort the computation after N seconds, will raise a failure exit code. Calls 'interruptProcessGroupOf' and 'terminateProcess', but may sometimes fail to abort the process and not timeout.
-    | WithStdout Bool -- ^ Should I include the @stdout@ in the exception if the command fails? Defaults to 'False'.
-    | WithStderr Bool -- ^ Should I include the @stderr@ in the exception if the command fails? Defaults to 'True'.
-    | EchoStdout Bool -- ^ Should I echo the @stdout@? Defaults to 'True' unless a 'Stdout' result is required or you use 'FileStdout'.
-    | EchoStderr Bool -- ^ Should I echo the @stderr@? Defaults to 'True' unless a 'Stderr' result is required or you use 'FileStderr'.
-    | FileStdout FilePath -- ^ Should I put the @stdout@ to a file.
-    | FileStderr FilePath -- ^ Should I put the @stderr@ to a file.
-      deriving (Eq,Ord,Show)
-
-
 -- | /Deprecated:/ Use 'AddPath'. This function will be removed in a future version.
 --
 --   Add a prefix and suffix to the @$PATH@ environment variable. For example:
@@ -100,7 +83,7 @@ addPath pre post = do
 addEnv :: MonadIO m => [(String, String)] -> m CmdOption
 addEnv extra = do
     args <- liftIO getEnvironment
-    return $ Env $ extra ++ filter (\(a,b) -> a `notElem` map fst extra) args
+    return $ Env $ extra ++ filter (\(a,_) -> a `notElem` map fst extra) args
 
 
 data Str = Str String | BS BS.ByteString | LBS LBS.ByteString | Unit deriving Eq
@@ -122,121 +105,105 @@ instance Eq Pid where _ == _ = True
 ---------------------------------------------------------------------
 -- ACTION EXPLICIT OPERATION
 
+-- | Given explicit operations, apply the advance ones, like skip/trace/track/autodep
 commandExplicit :: String -> [CmdOption] -> [Result] -> String -> [String] -> Action [Result]
-commandExplicit funcName copts results exe args = do
-    opts <- getShakeOptions
-    verb <- getVerbosity
+commandExplicit funcName oopts results exe args = do
+    ShakeOptions
+        {shakeCommandOptions,shakeRunCommands
+        ,shakeLint,shakeLintInside,shakeLintIgnore} <- getShakeOptions
+    let fopts = shakeCommandOptions ++ oopts
+    let useShell = Shell `elem` fopts
+    let useLint = shakeLint == Just LintFSATrace
+    let useAutoDeps = AutoDeps `elem` fopts
+    let opts = filter (/= Shell) fopts
 
-    let skipper act = if null results && not (shakeRunCommands opts) then return [] else act
+    let skipper act = if null results && not shakeRunCommands then return [] else act
 
     let verboser act = do
-            let cwd = listToMaybe $ reverse [x | Cwd x <- copts]
-            putLoud $ maybe "" (\x -> "cd " ++ x ++ "; ") cwd ++ saneCommandForUser exe args
+            let cwd = listToMaybe $ reverse [x | Cwd x <- opts]
+            putLoud $ maybe "" (\x -> "cd " ++ x ++ "; ") cwd ++ showCommandForUser2 exe args
+            verb <- getVerbosity
             (if verb >= Loud then quietly else id) act
 
-    let tracer = case reverse [x | Traced x <- copts] of
+    let tracer = case reverse [x | Traced x <- opts] of
             "":_ -> liftIO
             msg:_ -> traced msg
             [] -> traced (takeFileName exe)
 
-    let tracker act = case shakeLint opts of
-            Just LintTracker -> (if isWindows then winTracker else unixTracker) act
-            _ -> act [] exe args
+    let tracker act
+            | useLint = fsatrace act
+            | useAutoDeps = autodeps act
+            | useShell = shelled act
+            | otherwise = act exe args
 
-        winTracker act = do
-            (dir, cleanup) <- liftIO newTempDir
-            flip actionFinally cleanup $ do
-                res <- act [] "tracker" $ "/if":dir:"/c":exe:args
-                (rs, ws) <- liftIO $ trackerFiles dir
-                trackRead rs
-                trackWrite ws
-                return res
+        shelled = runShell (unwords $ exe : args)
+                              
+        ignore = map (?==) shakeLintIgnore
+        ham cwd xs = [makeRelative cwd x | x <- map toStandard xs
+                                         , any (`isPrefixOf` x) shakeLintInside
+                                         , not $ any ($ x) ignore]
 
-        unixTracker act = do
-            (file, cleanup) <- liftIO newTempFile
-            flip actionFinally cleanup $ do
-                fsat <- liftIO $ getEnv "FSAT"
-                let vars = [AddEnv "DYLD_INSERT_LIBRARIES" fsat
-                           ,AddEnv "DYLD_FORCE_FLAT_NAMESPACE" "1"
-                           ,AddEnv "FSAT_OUT" file]
-                res <- act vars exe args
-                (rs, ws) <- liftIO $ fsatraceFiles file
-                whitelist <- liftIO unixWhitelist
-                let whitelisted x = any (\w -> (w ++ "/") `isPrefixOf` x) whitelist
-                trackRead $ filter (not . whitelisted) rs
-                trackWrite $ filter (not . whitelisted) ws
-                return res
+        fsaCmd act opts file
+            | useShell = runShell (unwords $ exe : args) $ \exe args -> act "fsatrace" $ opts : file : "--" : exe : args
+            | otherwise = act "fsatrace" $ opts : file : "--" : exe : args
 
-    skipper $ tracker $ \opts exe args -> verboser $ tracer $ commandExplicitIO funcName (opts++copts) results exe args
+        fsatrace act = withTempFile $ \file -> do
+            res <- fsaCmd act "rwm" file
+            xs <- liftIO $ parseFSAT <$> readFileUTF8' file
+            cwd <- liftIO getCurrentDirectory
+            let reader (FSATRead x) = Just x; reader _ = Nothing
+                writer (FSATWrite x) = Just x; writer (FSATMove x _) = Just x; writer _ = Nothing
+                existing f = liftIO . filterM doesFileExist . nubOrd . mapMaybe f
+            rs <- existing reader xs
+            ws <- existing writer xs
+            let reads = ham cwd rs
+                writes = ham cwd ws
+            when useAutoDeps $
+                unsafeAllowApply $ needed reads
+            trackRead reads
+            trackWrite writes
+            return res
+
+        autodeps act = withTempFile $ \file -> do
+            res <-  fsaCmd act "r" file
+            pxs <- liftIO $ parseFSAT <$> readFileUTF8' file
+            xs <- liftIO $ filterM doesFileExist [x | FSATRead x <- pxs]
+            cwd <- liftIO getCurrentDirectory
+            unsafeAllowApply $ needNorm $ ham cwd xs
+            return res
+
+    skipper $ tracker $ \exe args -> verboser $ tracer $ commandExplicitIO funcName opts results exe args
 
 
--- | Given a directory (as passed to tracker /if) report on which files were used for reading/writing
-trackerFiles :: FilePath -> IO ([FilePath], [FilePath])
-trackerFiles dir = do
-    curdir <- getCurrentDirectory
-    let pre = upper curdir ++ "\\"
-    files <- getDirectoryContents dir
-    let f typ = do
-            files <- forM [x | x <- files, takeExtension x == ".tlog", takeExtension (dropExtension $ dropExtension x) == '.':typ] $ \file -> do
-                xs <- readFileEncoding utf16 $ dir </> file
-                return $ filter (not . isPrefixOf "." . takeFileName) . mapMaybe (stripPrefix pre) $ lines xs
-            fmap nubOrd $ mapMaybeM correctCase $ nubOrd $ concat files
-    liftM2 (,) (f "read") (f "write")
+-- | Given a shell command, call the continuation with the sanitised exec-style arguments
+runShell :: String -> (String -> [String] -> Action a) -> Action a
+runShell x act | not isWindows = act "/bin/sh" ["-c",x] -- do exactly what Haskell does
+runShell x act = withTempDir $ \dir -> do
+    let file = dir </> "s.bat"
+    writeFile' file x
+    act "cmd.exe" ["/d/q/c",file]
 
 
-correctCase :: FilePath -> IO (Maybe FilePath)
-correctCase x = f "" x
-    where
-        f pre "" = return $ Just pre
-        f pre x = do
-            let (a,b) = (takeDirectory1 x, dropDirectory1 x)
-            dir <- getDirectoryContents pre
-            case find ((==) a . upper) dir of
-                Nothing -> return Nothing -- if it can't be found it probably doesn't exist, so assume a file that wasn't really read
-                Just v -> f (pre +/+ v) b
+-- | Parse the FSATrace structure
+data FSAT
+    = FSATWrite FilePath
+    | FSATRead FilePath
+    | FSATDelete FilePath
+    | FSATMove FilePath FilePath
 
-        a +/+ b = if null a then b else a ++ "/" ++ b
-
-
-fsatraceFiles :: FilePath -> IO ([FilePath], [FilePath])
-fsatraceFiles file = do
-    xs <- parseFSAT <$> readFileUTF8 file
-    let reader (FSATRead x) = Just x; reader _ = Nothing
-        writer (FSATWrite x) = Just x; writer (FSATMove x y) = Just x; writer _ = Nothing
-    frs <- liftIO $ filterM doesFileExist $ nubOrd $ map normalise $ mapMaybe reader xs
-    fws <- liftIO $ filterM doesFileExist $ nubOrd $ map normalise $ mapMaybe writer xs
-    return (frs, fws)
-
-
-data FSAT = FSATWrite FilePath | FSATRead FilePath | FSATMove FilePath FilePath | FSATDelete FilePath
-
-parseFSAT :: String -> [FSAT] -- any parse errors are skipped
-parseFSAT = mapMaybe (f . wordsBy (== ':')) . lines
-    where f ["w",x] = Just $ FSATWrite x
-          f ["r",x] = Just $ FSATRead x
-          f ["m",x,y] = Just $ FSATMove x y
-          f ["d",x] = Just $ FSATDelete x
+-- | Parse the 'FSAT' entries, ignoring anything you don't understand.
+parseFSAT :: String -> [FSAT]
+parseFSAT = mapMaybe f . lines
+    where f ('w':'|':xs) = Just $ FSATWrite xs
+          f ('r':'|':xs) = Just $ FSATRead xs
+          f ('d':'|':xs) = Just $ FSATDelete xs
+          f ('m':'|':xs) | (xs,'|':ys) <- break (== '|') xs = Just $ FSATMove xs ys
           f _ = Nothing
 
-
-unixWhitelist :: IO [FilePath]
-unixWhitelist = do
-    home <- getEnv "HOME"
-    return [home ++ "/.ghc"
-           ,home ++ "/Library/Haskell"
-           ,home ++ "/Applications"
-           ,home ++ "/.cabal"
-           ,"/Applications"
-           ,"/var/folders"
-           ,"/usr"
-           ,"/Library"
-           ,"/System"
-           ]
-
-
 ---------------------------------------------------------------------
 -- IO EXPLICIT OPERATION
 
+-- | Given a very explicit set of CmdOption, translate them to a General.Process structure
 commandExplicitIO :: String -> [CmdOption] -> [Result] -> String -> [String] -> IO [Result]
 commandExplicitIO funcName opts results exe args = do
     let (grabStdout, grabStderr) = both or $ unzip $ for results $ \r -> case r of
@@ -247,7 +214,11 @@ commandExplicitIO funcName opts results exe args = do
 
     optEnv <- resolveEnv opts
     let optCwd = let x = last $ "" : [x | Cwd x <- opts] in if x == "" then Nothing else Just x
-    let optStdin = flip mapMaybe opts $ \x -> case x of Stdin x -> Just $ Left x; StdinBS x -> Just $ Right x; _ -> Nothing
+    let optStdin = flip mapMaybe opts $ \x -> case x of
+            Stdin x -> Just $ SrcString x
+            StdinBS x -> Just $ SrcBytes x
+            FileStdin x -> Just $ SrcFile x
+            _ -> Nothing
     let optShell = Shell `elem` opts
     let optBinary = BinaryPipes `elem` opts
     let optAsync = ResultProcess Pid0 `elem` results
@@ -259,7 +230,7 @@ commandExplicitIO funcName opts results exe args = do
     let optEchoStdout = last $ (not grabStdout && null optFileStdout) : [x | EchoStdout x <- opts]
     let optEchoStderr = last $ (not grabStderr && null optFileStderr) : [x | EchoStderr x <- opts]
 
-    let cmdline = saneCommandForUser exe args
+    let cmdline = showCommandForUser2 exe args
     let bufLBS f = do (a,b) <- buf $ LBS LBS.empty; return (a, (\(LBS x) -> f x) <$> b)
         buf Str{} | optBinary = bufLBS (Str . LBS.unpack)
         buf Str{} = do x <- newBuffer; return ([DestString x | not optAsync], Str . concat <$> readBuffer x)
@@ -268,19 +239,19 @@ commandExplicitIO funcName opts results exe args = do
         buf Unit  = return ([], return Unit)
     (dStdout, dStderr, resultBuild) :: ([[Destination]], [[Destination]], [Double -> ProcessHandle -> ExitCode -> IO Result]) <-
         fmap unzip3 $ forM results $ \r -> case r of
-            ResultCode _ -> return ([], [], \dur pid ex -> return $ ResultCode ex)
-            ResultTime _ -> return ([], [], \dur pid ex -> return $ ResultTime dur)
-            ResultLine _ -> return ([], [], \dur pid ex -> return $ ResultLine cmdline)
-            ResultProcess _ -> return ([], [], \dur pid ex -> return $ ResultProcess $ Pid pid)
+            ResultCode _ -> return ([], [], \_ _ ex -> return $ ResultCode ex)
+            ResultTime _ -> return ([], [], \dur _ _ -> return $ ResultTime dur)
+            ResultLine _ -> return ([], [], \_ _ _ -> return $ ResultLine cmdline)
+            ResultProcess _ -> return ([], [], \_ pid _ -> return $ ResultProcess $ Pid pid)
             ResultStdout    s -> do (a,b) <- buf s; return (a , [], \_ _ _ -> fmap ResultStdout b)
             ResultStderr    s -> do (a,b) <- buf s; return ([], a , \_ _ _ -> fmap ResultStderr b)
             ResultStdouterr s -> do (a,b) <- buf s; return (a , a , \_ _ _ -> fmap ResultStdouterr b)
 
     exceptionBuffer <- newBuffer
-    po <- resolvePath $ ProcessOpts
+    po <- resolvePath ProcessOpts
         {poCommand = if optShell then ShellCommand $ unwords $ exe:args else RawCommand exe args
         ,poCwd = optCwd, poEnv = optEnv, poTimeout = optTimeout
-        ,poStdin = if optBinary || any isRight optStdin then Right $ LBS.concat $ map (either LBS.pack id) optStdin else Left $ concatMap fromLeft optStdin
+        ,poStdin = [SrcBytes LBS.empty | optBinary && not (null optStdin)] ++ optStdin
         ,poStdout = [DestEcho | optEchoStdout] ++ map DestFile optFileStdout ++ [DestString exceptionBuffer | optWithStdout && not optAsync] ++ concat dStdout
         ,poStderr = [DestEcho | optEchoStderr] ++ map DestFile optFileStderr ++ [DestString exceptionBuffer | optWithStderr && not optAsync] ++ concat dStderr
         ,poAsync = optAsync
@@ -299,7 +270,7 @@ commandExplicitIO funcName opts results exe args = do
                 cwd ++ extra
     case res of
         Left err -> failure $ show err
-        Right (dur,(pid,ex)) | ex /= ExitSuccess && ResultCode ExitSuccess `notElem` results -> do
+        Right (_,(_,ex)) | ex /= ExitSuccess && ResultCode ExitSuccess `notElem` results -> do
             exceptionBuffer <- readBuffer exceptionBuffer
             let captured = ["Stderr" | optWithStderr] ++ ["Stdout" | optWithStdout]
             failure $
@@ -310,14 +281,16 @@ commandExplicitIO funcName opts results exe args = do
         Right (dur,(pid,ex)) -> mapM (\f -> f dur pid ex) resultBuild
 
 
+-- | Apply all environment operations, to produce a new environment to use.
 resolveEnv :: [CmdOption] -> IO (Maybe [(String, String)])
 resolveEnv opts
-    | null env, null addEnv, null addPath = return Nothing
-    | otherwise = Just . unique . tweakPath . (++ addEnv) <$>
+    | null env, null addEnv, null addPath, null remEnv = return Nothing
+    | otherwise = Just . unique . tweakPath . (++ addEnv) . filter (flip notElem remEnv . fst) <$>
                   if null env then getEnvironment else return (concat env)
     where
         env = [x | Env x <- opts]
         addEnv = [(x,y) | AddEnv x y <- opts]
+        remEnv = [x | RemEnv x <- opts]
         addPath = [(x,y) | AddPath x y <- opts]
 
         newPath mid = intercalate [searchPathSeparator] $
@@ -339,7 +312,7 @@ resolvePath po
     = do
     let progExe = if prog == prog -<.> exe then prog else prog <.> exe
     -- use unsafeInterleaveIO to allow laziness to skip the queries we don't use
-    pathOld <- unsafeInterleaveIO $ fmap (fromMaybe "") $ lookupEnv "PATH"
+    pathOld <- unsafeInterleaveIO $ fromMaybe "" <$> lookupEnv "PATH"
     old <- unsafeInterleaveIO $ findExecutable prog
     new <- unsafeInterleaveIO $ findExecutableWith (splitSearchPath path) progExe
     old2 <- unsafeInterleaveIO $ findExecutableWith (splitSearchPath pathOld) progExe
@@ -357,20 +330,13 @@ resolvePath po
 resolvePath po = return po
 
 
+-- | Given a list of directories, and a file name, return the complete path if you can find it.
+--   Like findExecutable, but with a custom PATH.
 findExecutableWith :: [FilePath] -> String -> IO (Maybe FilePath)
 findExecutableWith path x = flip firstJustM (map (</> x) path) $ \s ->
     ifM (doesFileExist s) (return $ Just s) (return Nothing)
 
 
--- Like System.Process, but tweaked to show less escaping,
--- Relies on relatively detailed internals of showCommandForUser.
-saneCommandForUser :: FilePath -> [String] -> String
-saneCommandForUser cmd args = unwords $ map f $ cmd:args
-    where
-        f x = if take (length y - 2) (drop 1 y) == x then x else y
-            where y = showCommandForUser x []
-
-
 ---------------------------------------------------------------------
 -- FIXED ARGUMENT WRAPPER
 
@@ -402,14 +368,14 @@ newtype Process = Process {fromProcess :: ProcessHandle}
 --   write helper functions that print out the time of a result.
 --
 -- @
---timer :: ('CmdResult' r, MonadIO m) => (forall r . 'CmdResult' r => m r) -> m r
---timer act = do
---    ('CmdTime' t, 'CmdLine' x, r) <- act
---    liftIO $ putStrLn $ \"Command \" ++ x ++ \" took \" ++ show t ++ \" seconds\"
---    return r
+-- timer :: ('CmdResult' r, MonadIO m) => (forall r . 'CmdResult' r => m r) -> m r
+-- timer act = do
+--     ('CmdTime' t, 'CmdLine' x, r) <- act
+--     liftIO $ putStrLn $ \"Command \" ++ x ++ \" took \" ++ show t ++ \" seconds\"
+--     return r
 --
---run :: IO ()
---run = timer $ 'cmd' \"ghc --version\"
+-- run :: IO ()
+-- run = timer $ 'cmd' \"ghc --version\"
 -- @
 newtype CmdTime = CmdTime {fromCmdTime :: Double}
 
@@ -465,6 +431,7 @@ instance (CmdResult x1, CmdResult x2) => CmdResult (x1,x2) where
         where (a1,b1) = cmdResult
               (a2,b2) = cmdResult
 
+cmdResultWith :: forall b c. CmdResult b => (b -> c) -> ([Result], [Result] -> c)
 cmdResultWith f = second (f .) cmdResult
 
 instance (CmdResult x1, CmdResult x2, CmdResult x3) => CmdResult (x1,x2,x3) where
@@ -504,7 +471,7 @@ instance (CmdResult x1, CmdResult x2, CmdResult x3, CmdResult x4, CmdResult x5)
 --   pass @'WithStderr' 'False'@, which causes no streams to be captured by Shake, and certain programs (e.g. @gcc@)
 --   to detect they are running in a terminal.
 command :: CmdResult r => [CmdOption] -> String -> [String] -> Action r
-command opts x xs = fmap b $ commandExplicit "command" opts a x xs
+command opts x xs = b <$> commandExplicit "command" opts a x xs
     where (a,b) = cmdResult
 
 -- | A version of 'command' where you do not require any results, used to avoid errors about being unable
@@ -516,6 +483,8 @@ command_ opts x xs = void $ commandExplicit "command_" opts [] x xs
 ---------------------------------------------------------------------
 -- VARIABLE ARGUMENT WRAPPER
 
+-- | A type annotation, equivalent to the first argument, but in variable argument contexts,
+--   gives a clue as to what return type is expected (not actually enforced).
 type a :-> t = a
 
 
@@ -569,16 +538,37 @@ instance (Arg a, CmdArguments r) => CmdArguments (a -> r) where
     cmdArguments xs x = cmdArguments $ xs ++ arg x
 instance CmdResult r => CmdArguments (Action r) where
     cmdArguments x = case partitionEithers x of
-        (opts, x:xs) -> let (a,b) = cmdResult in fmap b $ commandExplicit "cmd" opts a x xs
+        (opts, x:xs) -> let (a,b) = cmdResult in b <$> commandExplicit "cmd" opts a x xs
         _ -> error "Error, no executable or arguments given to Development.Shake.cmd"
 instance CmdResult r => CmdArguments (IO r) where
     cmdArguments x = case partitionEithers x of
-        (opts, x:xs) -> let (a,b) = cmdResult in fmap b $ commandExplicitIO "cmd" opts a x xs
+        (opts, x:xs) -> let (a,b) = cmdResult in b <$> commandExplicitIO "cmd" opts a x xs
         _ -> error "Error, no executable or arguments given to Development.Shake.cmd"
 
+instance CmdArguments [Either CmdOption String] where
+    cmdArguments = id
+
+
 class Arg a where arg :: a -> [Either CmdOption String]
 instance Arg String where arg = map Right . words
 instance Arg [String] where arg = map Right
 instance Arg CmdOption where arg = return . Left
 instance Arg [CmdOption] where arg = map Left
 instance Arg a => Arg (Maybe a) where arg = maybe [] arg
+
+
+---------------------------------------------------------------------
+-- UTILITIES
+
+-- Copied from Derived. Once Derived no longer exports cmd stuff, import from there.
+withTempFile :: (FilePath -> Action a) -> Action a
+withTempFile act = do
+    (file, del) <- liftIO newTempFile
+    act file `actionFinally` del
+
+-- A better version of showCommandForUser, which doesn't escape so much on Windows
+showCommandForUser2 :: FilePath -> [String] -> String
+showCommandForUser2 cmd args = unwords $ map (\x -> if safe x then x else showCommandForUser x []) $ cmd : args
+    where
+        safe xs = not (null xs) && not (any bad xs)
+        bad x = isSpace x || (x == '\\' && not isWindows) || x `elem` "\"\'"
