@@ -1,25 +1,20 @@
 {-# LANGUAGE RecordWildCards, GeneralizedNewtypeDeriving, ScopedTypeVariables, PatternGuards #-}
-{-# LANGUAGE ExistentialQuantification, MultiParamTypeClasses #-}
-
-{-# LANGUAGE CPP #-}
-#if __GLASGOW_HASKELL__ >= 704
-{-# LANGUAGE ConstraintKinds #-}
-#endif
+{-# LANGUAGE ExistentialQuantification, MultiParamTypeClasses, ConstraintKinds #-}
 
 module Development.Shake.Core(
     run,
-#if __GLASGOW_HASKELL__ >= 704
     ShakeValue,
-#endif
     Rule(..), Rules, rule, action, withoutActions, alternatives, priority,
-    Action, actionOnException, actionFinally, apply, apply1, traced, getShakeOptions,
+    Action, actionOnException, actionFinally, apply, apply1, traced, getShakeOptions, getProgress,
     trackUse, trackChange, trackAllow,
     getVerbosity, putLoud, putNormal, putQuiet, withVerbosity, quietly,
     Resource, newResource, newResourceIO, withResource, withResources, newThrottle, newThrottleIO,
     newCache, newCacheIO,
-    unsafeExtraThread,
+    unsafeExtraThread, unsafeAllowApply,
+    parallel,
+    orderOnlyAction,
     -- Internal stuff
-    rulesIO, runAfter, unsafeIgnoreDependencies,
+    runAfter
     ) where
 
 import Control.Exception.Extra
@@ -27,13 +22,14 @@ import Control.Applicative
 import Data.Tuple.Extra
 import Control.Concurrent.Extra
 import Control.Monad.Extra
+import Control.Monad.Fix
 import Control.Monad.IO.Class
 import Control.Monad.Trans.Writer.Strict
 import Data.Typeable
 import Data.Function
 import Data.Either.Extra
 import Numeric.Extra
-import Data.List
+import Data.List.Extra
 import qualified Data.HashMap.Strict as Map
 import Data.Maybe
 import Data.IORef
@@ -64,18 +60,6 @@ import Prelude
 ---------------------------------------------------------------------
 -- RULES
 
-#if __GLASGOW_HASKELL__ >= 704
--- | Define an alias for the six type classes required for things involved in Shake 'Development.Shake.Rule's.
---   This alias is only available in GHC 7.4 and above, and requires the @ConstraintKinds@ extension.
---
---   To define your own values meeting the necessary constraints it is convenient to use the extensions
---   @GeneralizedNewtypeDeriving@ and @DeriveDataTypeable@ to write:
---
--- > newtype MyType = MyType (String, Bool) deriving (Show,Typeable,Eq,Hashable,Binary,NFData)
-type ShakeValue a = (Show a, Typeable a, Eq a, Hashable a, Binary a, NFData a)
-#endif
-
-
 -- | Define a pair of types that can be used by Shake rules.
 --   To import all the type classes required see "Development.Shake.Classes".
 --
@@ -102,16 +86,16 @@ type ShakeValue a = (Show a, Typeable a, Eq a, Hashable a, Binary a, NFData a)
 --   provide a similar rule).
 --
 -- @
---newtype File = File FilePath deriving (Show, Typeable, Eq, Hashable, Binary, NFData)
---newtype Modtime = Modtime Double deriving (Show, Typeable, Eq, Hashable, Binary, NFData)
---getFileModtime file = ...
+-- newtype File = File FilePath deriving (Show, Typeable, Eq, Hashable, Binary, NFData)
+-- newtype Modtime = Modtime Double deriving (Show, Typeable, Eq, Hashable, Binary, NFData)
+-- getFileModtime file = ...
 --
---instance Rule File Modtime where
---    storedValue _ (File x) = do
---        exists <- System.Directory.doesFileExist x
---        if exists then Just \<$\> getFileModtime x else return Nothing
---    equalValue _ _ t1 t2 =
---        if t1 == t2 then EqualCheap else NotEqual
+-- instance Rule File Modtime where
+--     storedValue _ (File x) = do
+--         exists <- System.Directory.doesFileExist x
+--         if exists then Just \<$\> getFileModtime x else return Nothing
+--     equalValue _ _ t1 t2 =
+--         if t1 == t2 then EqualCheap else NotEqual
 -- @
 --
 --   This example instance means:
@@ -150,14 +134,7 @@ type ShakeValue a = (Show a, Typeable a, Eq a, Hashable a, Binary a, NFData a)
 --   For rules whose values are not stored externally,
 --   'storedValue' should return 'Just' with a sentinel value
 --   and 'equalValue' should always return 'EqualCheap' for that sentinel.
-class (
-#if __GLASGOW_HASKELL__ >= 704
-    ShakeValue key, ShakeValue value
-#else
-    Show key, Typeable key, Eq key, Hashable key, Binary key, NFData key,
-    Show value, Typeable value, Eq value, Hashable value, Binary value, NFData value
-#endif
-    ) => Rule key value where
+class (ShakeValue key, ShakeValue value) => Rule key value where
 
     -- | /[Required]/ Retrieve the @value@ associated with a @key@, if available.
     --
@@ -179,14 +156,11 @@ ruleValue :: (key -> Maybe (m value)) -> value
 ruleValue = err "ruleValue"
 
 
--- | Define a set of rules. Rules can be created with calls to functions such as 'Development.Shake.%>' or 'action'. Rules are combined
---   with either the 'Monoid' instance, or (more commonly) the 'Monad' instance and @do@ notation. To define your own
---   custom types of rule, see "Development.Shake.Rule".
+-- | Define a set of rules. Rules can be created with calls to functions such as 'Development.Shake.%>' or 'action'.
+--   Rules are combined with either the 'Monoid' instance, or (more commonly) the 'Monad' instance and @do@ notation.
+--   To define your own custom types of rule, see "Development.Shake.Rule".
 newtype Rules a = Rules (WriterT (SRules Action) IO a) -- All IO must be associative/commutative (e.g. creating IORef/MVars)
-    deriving (Monad, Functor, Applicative)
-
-rulesIO :: IO a -> Rules a
-rulesIO = Rules . liftIO
+    deriving (Functor, Applicative, Monad, MonadIO, MonadFix)
 
 newRules :: SRules Action -> Rules ()
 newRules = Rules . tell
@@ -215,8 +189,10 @@ instance Monoid a => Monoid (Rules a) where
     mappend = liftA2 mappend
 
 
--- | Add a rule to build a key, returning an appropriate 'Action'. All rules at a given priority
---   must be disjoint. Rules have priority 1 by default, but can be modified with 'priority'.
+-- | Add a rule to build a key, returning an appropriate 'Action' if the @key@ matches,
+--   or 'Nothing' otherwise.
+--   All rules at a given priority must be disjoint on all used @key@ values, with at most one match.
+--   Rules have priority 1 by default, which can be modified with 'priority'.
 rule :: Rule key value => (key -> Maybe (Action value)) -> Rules ()
 rule r = newRules mempty{rules = Map.singleton k (k, v, [(1,ARule r)])}
     where k = typeOf $ ruleKey r; v = typeOf $ ruleValue r
@@ -310,10 +286,10 @@ createRuleinfo opt SRules{..} = flip Map.map rules $ \(_,tv,rs) -> RuleInfo (sto
         execute rs = \k -> case filter (not . null) $ map (mapMaybe ($ k)) rs2 of
                [r]:_ -> r
                rs -> liftIO $ errorMultipleRulesMatch (typeKey k) (show k) (length rs)
-            where rs2 = sets [(i, \k -> fmap (fmap newValue) $ r (fromKey k)) | (i,ARule r) <- rs]
+            where rs2 = sets [(i, \k -> fmap newValue <$> r (fromKey k)) | (i,ARule r) <- rs]
 
         sets :: Ord a => [(a, b)] -> [[b]] -- highest to lowest
-        sets = map (map snd) . reverse . groupBy ((==) `on` fst) . sortBy (compare `on` fst)
+        sets = map snd . reverse . groupSort
 
 runStored :: Map.HashMap TypeRep (RuleInfo m) -> Key -> IO (Maybe Value)
 runStored mp k = case Map.lookup (typeKey k) mp of
@@ -347,6 +323,7 @@ data Global = Global
     ,globalLint :: String -> IO ()
     ,globalAfter :: IORef [IO ()]
     ,globalTrackAbsent :: IORef [(Key, Key)] -- in rule fst, snd must be absent
+    ,globalProgress :: IO Progress
     }
 
 
@@ -403,6 +380,7 @@ run opts@ShakeOptions{..} rs = (if shakeLineBuffering then lineBuffering else id
 
     let diagnostic = if shakeVerbosity >= Diagnostic then outputLocked Diagnostic . ("% "++) else const $ return ()
     let output v = outputLocked v . abbreviate shakeAbbreviations
+    diagnostic "Starting run"
 
     except <- newIORef (Nothing :: Maybe (String, ShakeException))
     let raiseError err
@@ -422,6 +400,7 @@ run opts@ShakeOptions{..} rs = (if shakeLineBuffering then lineBuffering else id
                 ,("Wanted",Just dir)
                 ,("Got",Just now)]
                 ""
+    diagnostic "Starting run 2"
 
     after <- newIORef []
     absent <- newIORef []
@@ -430,29 +409,32 @@ run opts@ShakeOptions{..} rs = (if shakeLineBuffering then lineBuffering else id
             when shakeTimings printTimings
             resetTimings -- so we don't leak memory
         withNumCapabilities shakeThreads $ do
+            diagnostic "Starting run 3"
             withDatabase opts diagnostic $ \database -> do
                 wait <- newBarrier
-                tid <- forkIO $ flip finally (signalBarrier wait ()) $
-                    shakeProgress $ do
-                        failure <- fmap (fmap fst) $ readIORef except
+                let getProgress = do
+                        failure <- fmap fst <$> readIORef except
                         stats <- progress database
                         return stats{isFailure=failure}
-                addCleanup cleanup $ do
+                tid <- flip forkFinally (const $ signalBarrier wait ()) $
+                    shakeProgress getProgress
+                _ <- addCleanup cleanup $ do
                     killThread tid
                     void $ timeout 1000000 $ waitBarrier wait
 
                 let ruleinfo = createRuleinfo opts rs
                 addTiming "Running rules"
                 runPool (shakeThreads == 1) shakeThreads $ \pool -> do
-                    let s0 = Global database pool cleanup start ruleinfo output opts diagnostic lint after absent
+                    let s0 = Global database pool cleanup start ruleinfo output opts diagnostic lint after absent getProgress
                     let s1 = Local emptyStack shakeVerbosity Nothing [] 0 [] [] []
-                    forM_ (actions rs) $ \act -> do
+                    forM_ (actions rs) $ \act ->
                         addPool pool $ runAction s0 s1 act $ \x -> case x of
                             Left e -> raiseError =<< shakeException s0 (return ["Top-level action/want"]) e
                             Right x -> return x
                 maybe (return ()) (throwIO . snd) =<< readIORef except
+                assertFinishedDatabase database
 
-                when (null $ actions rs) $ do
+                when (null $ actions rs) $
                     when (shakeVerbosity >= Normal) $ output Normal "Warning: No want/action statements, nothing to do"
 
                 when (isJust shakeLint) $ do
@@ -486,8 +468,8 @@ abbreviate :: [(String,String)] -> String -> String
 abbreviate [] = id
 abbreviate abbrev = f
     where
-        -- order so longer appreviations are preferred
-        ordAbbrev = sortBy (flip (compare `on` length . fst)) abbrev
+        -- order so longer abbreviations are preferred
+        ordAbbrev = sortOn (negate . length . fst) abbrev
 
         f [] = []
         f x | (to,rest):_ <- [(to,rest) | (from,to) <- ordAbbrev, Just rest <- [stripPrefix from x]] = to ++ f rest
@@ -495,7 +477,7 @@ abbreviate abbrev = f
 
 
 runAction :: Global -> Local -> Action a -> Capture (Either SomeException a)
-runAction g l (Action x) k = runRAW g l x k
+runAction g l (Action x) = runRAW g l x
 
 
 runAfter :: IO () -> Action ()
@@ -508,20 +490,21 @@ runAfter op = do
 --   This function requires that appropriate rules have been added with 'rule'.
 --   All @key@ values passed to 'apply' become dependencies of the 'Action'.
 apply :: Rule key value => [key] -> Action [value]
-apply = f -- Don't short-circuit [] as we still want error messages
-    where
-        -- We don't want the forall in the Haddock docs
-        f :: forall key value . Rule key value => [key] -> Action [value]
-        f ks = do
-            let tk = typeOf (err "apply key" :: key)
-                tv = typeOf (err "apply type" :: value)
-            Global{..} <- Action getRO
-            block <- Action $ getsRW localBlockApply
-            whenJust block $ liftIO . errorNoApply tk (fmap show $ listToMaybe ks)
-            case Map.lookup tk globalRules of
-                Nothing -> liftIO $ errorNoRuleToBuildType tk (fmap show $ listToMaybe ks) (Just tv)
-                Just RuleInfo{resultType=tv2} | tv /= tv2 -> liftIO $ errorRuleTypeMismatch tk (fmap show $ listToMaybe ks) tv2 tv
-                _ -> fmap (map fromValue) $ applyKeyValue $ map newKey ks
+apply = applyForall
+
+-- We don't want the forall in the Haddock docs
+-- Don't short-circuit [] as we still want error messages
+applyForall :: forall key value . Rule key value => [key] -> Action [value]
+applyForall ks = do
+    let tk = typeOf (err "apply key" :: key)
+        tv = typeOf (err "apply type" :: value)
+    Global{..} <- Action getRO
+    block <- Action $ getsRW localBlockApply
+    whenJust block $ liftIO . errorNoApply tk (show <$> listToMaybe ks)
+    case Map.lookup tk globalRules of
+        Nothing -> liftIO $ errorNoRuleToBuildType tk (show <$> listToMaybe ks) (Just tv)
+        Just RuleInfo{resultType=tv2} | tv /= tv2 -> liftIO $ errorRuleTypeMismatch tk (show <$> listToMaybe ks) tv2 tv
+        _ -> fmap (map fromValue) $ applyKeyValue $ map newKey ks
 
 
 applyKeyValue :: [Key] -> Action [Value]
@@ -538,8 +521,7 @@ applyKeyValue ks = do
                 liftIO $ globalLint $ "before building " ++ top
                 putWhen Chatty $ "# " ++ show k
                 res <- runExecute globalRules k
-                when (shakeLint globalOptions == Just LintTracker)
-                    trackCheckUsed
+                when (Just LintFSATrace == shakeLint globalOptions) trackCheckUsed
                 Action $ fmap ((,) res) getRW) $ \x -> case x of
                     Left e -> continue . Left . toException =<< shakeException global (showStack globalDatabase stack) e
                     Right (res, Local{..}) -> do
@@ -578,10 +560,24 @@ apply1 = fmap head . apply . return
 getShakeOptions :: Action ShakeOptions
 getShakeOptions = Action $ getsRO globalOptions
 
+-- | Get the current 'Progress' structure, as would be returned by 'shakeProgress'.
+getProgress :: Action Progress
+getProgress = do
+    res <- Action $ getsRO globalProgress
+    liftIO res
+
 
 -- | Write an action to the trace list, along with the start/end time of running the IO action.
 --   The 'Development.Shake.cmd' and 'Development.Shake.command' functions automatically call 'traced'.
 --   The trace list is used for profile reports (see 'shakeReport').
+--
+--   By default 'traced' prints some useful extra context about what
+--   Shake is building, e.g.:
+--
+-- > # traced message (for myobject.o)
+--
+--   To suppress the output of 'traced' (for example you want more control
+--   over the message using 'putNormal'), use the 'quietly' combinator.
 traced :: String -> IO a -> Action a
 traced msg act = do
     Global{..} <- Action getRO
@@ -602,12 +598,19 @@ putWhen v msg = do
         liftIO $ globalOutput v msg
 
 
--- | Write a message to the output when the verbosity ('shakeVerbosity') is appropriate.
---   The output will not be interleaved with any other Shake messages
---   (other than those generated by system commands).
-putLoud, putNormal, putQuiet :: String -> Action ()
+-- | Write an unimportant message to the output, only shown when 'shakeVerbosity' is higher than normal ('Loud' or above).
+--   The output will not be interleaved with any other Shake messages (other than those generated by system commands).
+putLoud :: String -> Action ()
 putLoud = putWhen Loud
+
+-- | Write a normal priority message to the output, only supressed when 'shakeVerbosity' is 'Quiet' or 'Silent'.
+--   The output will not be interleaved with any other Shake messages (other than those generated by system commands).
+putNormal :: String -> Action ()
 putNormal = putWhen Normal
+
+-- | Write an important message to the output, only supressed when 'shakeVerbosity' is 'Silent'.
+--   The output will not be interleaved with any other Shake messages (other than those generated by system commands).
+putQuiet :: String -> Action ()
 putQuiet = putWhen Quiet
 
 
@@ -639,13 +642,7 @@ quietly = withVerbosity Quiet
 -- TRACKING
 
 -- | Track that a key has been used by the action preceeding it.
-trackUse ::
-#if __GLASGOW_HASKELL__ >= 704
-    ShakeValue key
-#else
-    (Show key, Typeable key, Eq key, Hashable key, Binary key, NFData key)
-#endif
-    => key -> Action ()
+trackUse :: ShakeValue key => key -> Action ()
 -- One of the following must be true:
 -- 1) you are the one building this key (e.g. key == topStack)
 -- 2) you have already been used by apply, and are on the dependency list
@@ -684,7 +681,7 @@ trackCheckUsed = do
                 ""
 
         -- check 3b
-        bad <- flip filterM localTrackUsed $ \k -> fmap (not . null) $ lookupDependencies globalDatabase k
+        bad <- flip filterM localTrackUsed $ \k -> (not . null) <$> lookupDependencies globalDatabase k
         unless (null bad) $ do
             let n = length bad
             errorStructured
@@ -694,13 +691,7 @@ trackCheckUsed = do
 
 
 -- | Track that a key has been changed by the action preceeding it.
-trackChange ::
-#if __GLASGOW_HASKELL__ >= 704
-    ShakeValue key
-#else
-    (Show key, Typeable key, Eq key, Hashable key, Binary key, NFData key)
-#endif
-    => key -> Action ()
+trackChange :: ShakeValue key => key -> Action ()
 -- One of the following must be true:
 -- 1) you are the one building this key (e.g. key == topStack)
 -- 2) someone explicitly gave you permission with trackAllow
@@ -721,21 +712,15 @@ trackChange key = do
 
 
 -- | Allow any matching key to violate the tracking rules.
-trackAllow ::
-#if __GLASGOW_HASKELL__ >= 704
-    ShakeValue key
-#else
-    (Show key, Typeable key, Eq key, Hashable key, Binary key, NFData key)
-#endif
-    => (key -> Bool) -> Action ()
-trackAllow test = Action $ modifyRW $ \s -> s{localTrackAllows = f : localTrackAllows s}
-    where
-        -- We don't want the forall in the Haddock docs
-        arrow1Type :: forall a b . Typeable a => (a -> b) -> TypeRep
-        arrow1Type _ = typeOf (err "trackAllow" :: a)
+trackAllow :: ShakeValue key => (key -> Bool) -> Action ()
+trackAllow = trackAllowForall
 
-        ty = arrow1Type test
-        f k = typeKey k == ty && test (fromKey k)
+-- We don't want the forall in the Haddock docs
+trackAllowForall :: forall key . ShakeValue key => (key -> Bool) -> Action ()
+trackAllowForall test = Action $ modifyRW $ \s -> s{localTrackAllows = f : localTrackAllows s}
+    where
+        tk = typeOf (err "trackAllow key" :: key)
+        f k = typeKey k == tk && test (fromKey k)
 
 
 ---------------------------------------------------------------------
@@ -771,7 +756,7 @@ trackAllow test = Action $ modifyRW $ \s -> s{localTrackAllows = f : localTrackA
 --     'Development.Shake.cmd' \"cl -o\" [out] ...
 -- @
 newResource :: String -> Int -> Rules Resource
-newResource name mx = rulesIO $ newResourceIO name mx
+newResource name mx = liftIO $ newResourceIO name mx
 
 
 -- | Create a throttled resource, given a name (for error messages) and a number of resources (the 'Int') that can be
@@ -804,13 +789,17 @@ newResource name mx = rulesIO $ newResourceIO name mx
 --   we will never exceed an average of 1 request every 5 seconds, we may end up running an unbounded number of
 --   requests simultaneously. If this limitation causes a problem in practice it can be fixed.
 newThrottle :: String -> Int -> Double -> Rules Resource
-newThrottle name count period = rulesIO $ newThrottleIO name count period
+newThrottle name count period = liftIO $ newThrottleIO name count period
 
+unsafeAllowApply :: Action a -> Action a
+unsafeAllowApply  = applyBlockedBy Nothing
 
 blockApply :: String -> Action a -> Action a
-blockApply msg = Action . unmodifyRW f . fromAction
-    where f s0 = (s0{localBlockApply=Just msg}, \s -> s{localBlockApply=localBlockApply s0})
+blockApply = applyBlockedBy . Just
 
+applyBlockedBy :: Maybe String -> Action a -> Action a
+applyBlockedBy reason = Action . unmodifyRW f . fromAction
+    where f s0 = (s0{localBlockApply=reason}, \s -> s{localBlockApply=localBlockApply s0})
 
 -- | Run an action which uses part of a finite resource. For more details see 'Resource'.
 --   You cannot depend on a rule (e.g. 'need') while a resource is held.
@@ -818,7 +807,7 @@ withResource :: Resource -> Int -> Action a -> Action a
 withResource r i act = do
     Global{..} <- Action getRO
     liftIO $ globalDiagnostic $ show r ++ " waiting to acquire " ++ show i
-    offset <- liftIO $ offsetTime
+    offset <- liftIO offsetTime
     Action $ captureRAW $ \continue -> acquireResource r globalPool i $ continue $ Right ()
     res <- Action $ tryRAW $ fromAction $ blockApply ("Within withResource using " ++ show r) $ do
         offset <- liftIO offset
@@ -847,20 +836,21 @@ withResources res act
 newCacheIO :: (Eq k, Hashable k) => (k -> Action v) -> IO (k -> Action v)
 newCacheIO act = do
     var {- :: Var (Map k (Fence (Either SomeException ([Depends],v)))) -} <- newVar Map.empty
-    return $ \key -> do
+    return $ \key ->
         join $ liftIO $ modifyVar var $ \mp -> case Map.lookup key mp of
             Just bar -> return $ (,) mp $ do
                 res <- liftIO $ testFence bar
-                res <- case res of
-                    Just res -> return res
+                (res,offset) <- case res of
+                    Just res -> return (res, 0)
                     Nothing -> do
                         pool <- Action $ getsRO globalPool
+                        offset <- liftIO offsetTime
                         Action $ captureRAW $ \k -> waitFence bar $ \v ->
-                            addPool pool $ k $ Right v
+                            addPool pool $ do offset <- liftIO offset; k $ Right (v,offset)
                 case res of
                     Left err -> Action $ throwRAW err
                     Right (deps,v) -> do
-                        Action $ modifyRW $ \s -> s{localDepends = deps ++ localDepends s}
+                        Action $ modifyRW $ \s -> s{localDepends = deps ++ localDepends s, localDiscount = localDiscount s + offset}
                         return v
             Nothing -> do
                 bar <- newFence
@@ -898,26 +888,55 @@ newCacheIO act = do
 --   To create the result @MyFile.txt.digits@ the file @MyFile.txt@ will be read and counted, but only at most
 --   once per execution.
 newCache :: (Eq k, Hashable k) => (k -> Action v) -> Rules (k -> Action v)
-newCache = rulesIO . newCacheIO
+newCache = liftIO . newCacheIO
 
 
 -- | Run an action without counting to the thread limit, typically used for actions that execute
---   on remote machines using barely any local CPU resources. Unsafe as it allows the 'shakeThreads' limit to be exceeded.
+--   on remote machines using barely any local CPU resources.
+--   Unsafe as it allows the 'shakeThreads' limit to be exceeded.
 --   You cannot depend on a rule (e.g. 'need') while the extra thread is executing.
 --   If the rule blocks (e.g. calls 'withResource') then the extra thread may be used by some other action.
 --   Only really suitable for calling 'cmd'/'command'.
 unsafeExtraThread :: Action a -> Action a
 unsafeExtraThread act = Action $ do
-    global@Global{..} <- getRO
+    Global{..} <- getRO
     stop <- liftIO $ increasePool globalPool
     res <- tryRAW $ fromAction $ blockApply "Within unsafeExtraThread" act
     liftIO stop
     captureRAW $ \continue -> (if isLeft res then addPoolPriority else addPool) globalPool $ continue res
 
 
--- | Ignore any dependencies added by an action.
-unsafeIgnoreDependencies :: Action a -> Action a
-unsafeIgnoreDependencies act = Action $ do
+-- | Execute a list of actions in parallel. In most cases 'need' will be more appropriate to benefit from parallelism.
+parallel :: [Action a] -> Action [a]
+parallel [] = return []
+parallel [x] = fmap return x
+parallel acts = Action $ do
+    global@Global{..} <- getRO
+    local <- getRW
+    -- number of items still to complete, or Nothing for has completed (by either failure or completion)
+    todo :: Var (Maybe Int) <- liftIO $ newVar $ Just $ length acts
+    -- a list of refs where the results go
+    results :: [IORef (Maybe (Either SomeException a))] <- liftIO $ replicateM (length acts) $ newIORef Nothing
+
+    captureRAW $ \continue -> do
+        let resume = do
+                res <- liftIO $ sequence . catMaybes <$> mapM readIORef results
+                continue res
+
+        liftIO $ forM_ (zip acts results) $ \(act, result) -> do
+            let act2 = ifM (liftIO $ isJust <$> readVar todo) act (fail "")
+            addPool globalPool $ runAction global local act2 $ \res -> do
+                writeIORef result $ Just res
+                modifyVar_ todo $ \v -> case v of
+                    Nothing -> return Nothing
+                    Just i | i == 1 || isLeft res -> do resume; return Nothing
+                    Just i -> return $ Just $ i - 1
+
+
+-- | Run an action but do not depend on anything the action uses.
+--   A more general version of 'orderOnly'.
+orderOnlyAction :: Action a -> Action a
+orderOnlyAction act = Action $ do
     pre <- getsRW localDepends
     res <- fromAction act
     modifyRW $ \s -> s{localDepends=pre}
