@@ -4,7 +4,7 @@
 
 module Development.Shake.Database(
     Trace(..),
-    Database, withDatabase,
+    Database, withDatabase, assertFinishedDatabase,
     listDepends, lookupDependencies,
     Ops(..), build, Depends,
     progress,
@@ -110,7 +110,7 @@ data Result = Result
     {result :: Value -- the result associated with the Key
     ,built :: {-# UNPACK #-} !Step -- when it was actually run
     ,changed :: {-# UNPACK #-} !Step -- the step for deciding if it's valid
-    ,depends :: [[Id]] -- dependencies
+    ,depends :: [[Id]] -- dependencies (outer list is a strict ordered series, inner list is parallel)
     ,execution :: {-# UNPACK #-} !Float -- how long it took when it was last run (seconds)
     ,traces :: [Trace] -- a trace of the expensive operations (start/end in seconds since beginning of run)
     } deriving Show
@@ -185,7 +185,7 @@ data Ops = Ops
 
 -- | Return either an exception (crash), or (how much time you spent waiting, the value)
 build :: Pool -> Database -> Ops -> Stack -> [Key] -> Capture (Either SomeException (Seconds,Depends,[Value]))
-build pool Database{..} Ops{..} stack ks continue = do
+build pool database@Database{..} Ops{..} stack ks continue =
     join $ withLock lock $ do
         is <- forM ks $ \k -> do
             is <- readIORef intern
@@ -198,10 +198,15 @@ build pool Database{..} Ops{..} stack ks continue = do
                     return i
 
         whenJust (checkStack is stack) $ \bad -> do
+            -- everything else gets thrown via Left and can be Staunch'd
+            -- recursion in the rules is considered a worse error, so fails immediately
             status <- readIORef status
-            uncurry errorRuleRecursion $ case Map.lookup bad status of
+            let Stack _ xs _ = stack
+            stack <- return $ reverse $ map (maybe "<unknown>" (show . fst) . flip Map.lookup status) $ bad:xs
+            (tk, tname) <- return $ case Map.lookup bad status of
                 Nothing -> (Nothing, Nothing)
                 Just (k,_) -> (Just $ typeKey k, Just $ show k)
+            errorRuleRecursion stack tk tname
 
         vs <- mapM (reduce stack) is
         let errs = [e | Error e <- vs]
@@ -270,7 +275,7 @@ build pool Database{..} Ops{..} stack ks continue = do
         run stack i k r = do
             w <- newWaiting r
             addPool pool $ do
-                let reply res =  do
+                let reply res = do
                         ans <- withLock lock $ do
                             ans <- i #= (k, res)
                             runWaiting w
@@ -279,7 +284,7 @@ build pool Database{..} Ops{..} stack ks continue = do
                             Ready r -> do
                                 diagnostic $ "result " ++ atom k ++ " = "++ atom (result r) ++
                                              " " ++ (if built r == changed r then "(changed)" else "(unchanged)")
-                                journal i (k, Loaded r) -- leave the DB lock before appending
+                                journal i (k, Loaded r) -- we leave the DB lock before appending
                             Error _ -> do
                                 diagnostic $ "result " ++ atom k ++ " = error"
                                 journal i (k, Missing)
@@ -358,6 +363,15 @@ progress Database{..} = do
 ---------------------------------------------------------------------
 -- QUERY DATABASE
 
+assertFinishedDatabase :: Database -> IO ()
+assertFinishedDatabase Database{..} = do
+    -- if you have anyone Waiting, and are not exiting with an error, then must have a complex recursion (see #400)
+    status <- readIORef status
+    let bad = [key | (_, (key, Waiting{})) <- Map.toList status]
+    when (bad /= []) $
+        errorComplexRecursion (map show bad)
+
+
 -- | Given a map of representing a dependency order (with a show for error messages), find an ordering for the items such
 --   that no item points to an item before itself.
 --   Raise an error if you end up with a cycle.
@@ -378,7 +392,7 @@ dependencyOrder shw status = f (map fst noDeps) $ Map.map Just $ Map.fromListWit
                     "Internal invariant broken, database seems to be cyclic" :
                     map ("    " ++) bad ++
                     ["... plus " ++ show (length badOverflow) ++ " more ..." | not $ null badOverflow]
-            where (bad,badOverflow) = splitAt 10 $ [shw i | (i, Just _) <- Map.toList mp]
+            where (bad,badOverflow) = splitAt 10 [shw i | (i, Just _) <- Map.toList mp]
 
         f (x:xs) mp = x : f (now++xs) later
             where Just free = Map.lookupDefault (Just []) x mp
@@ -400,7 +414,7 @@ removeStep = Map.filter (\(k,_) -> k /= stepKey)
 
 toReport :: Database -> IO [ProfileEntry]
 toReport Database{..} = do
-    status <- fmap (removeStep . resultsOnly) $ readIORef status
+    status <- (removeStep . resultsOnly) <$> readIORef status
     let order = let shw i = maybe "<unknown>" (show . fst) $ Map.lookup i status
                 in dependencyOrder shw $ Map.map (concat . depends . snd) status
         ids = Map.fromList $ zip order [0..]
@@ -468,7 +482,7 @@ listDepends Database{..} (Depends xs) =
         return $ map (fst . fromJust . flip Map.lookup status) xs
 
 lookupDependencies :: Database -> Key -> IO [Key]
-lookupDependencies Database{..} k = do
+lookupDependencies Database{..} k =
     withLock lock $ do
         intern <- readIORef intern
         status <- readIORef status
@@ -531,4 +545,4 @@ instance BinaryWith Witness Status where
     putWith ctx Missing = putWord8 0
     putWith ctx (Loaded x) = putWord8 1 >> putWith ctx x
     putWith ctx x = err $ "putWith, Cannot write Status with constructor " ++ statusType x
-    getWith ctx = do i <- getWord8; if i == 0 then return Missing else fmap Loaded $ getWith ctx
+    getWith ctx = do i <- getWord8; if i == 0 then return Missing else Loaded <$> getWith ctx
