@@ -5,7 +5,7 @@ module Development.Shake.Errors(
     ShakeException(..),
     errorStructured, err,
     errorNoRuleToBuildType, errorRuleTypeMismatch, errorIncompatibleRules,
-    errorMultipleRulesMatch, errorRuleRecursion, errorNoApply,
+    errorMultipleRulesMatch, errorRuleRecursion, errorComplexRecursion, errorNoApply,
     ) where
 
 import Data.Tuple.Extra
@@ -29,7 +29,10 @@ alternatives = let (*) = (,) in
 
 
 errorStructured :: String -> [(String, Maybe String)] -> String -> IO a
-errorStructured msg args hint = errorIO $ unlines $
+errorStructured msg args hint = errorIO $ errorStructuredContents msg args hint
+
+errorStructuredContents :: String -> [(String, Maybe String)] -> String -> String
+errorStructuredContents msg args hint = unlines $
         [msg ++ ":"] ++
         ["  " ++ a ++ [':' | a /= ""] ++ replicate (as - length a + 2) ' ' ++ b | (a,b) <- args2] ++
         [hint | hint /= ""]
@@ -84,12 +87,21 @@ errorMultipleRulesMatch tk k count
     (if count == 0 then "Either add a rule that produces the above key, or stop requiring the above key"
      else "Modify your rules/defaultRules so only one can produce the above key")
 
-errorRuleRecursion :: Maybe TypeRep -> Maybe String -> IO a
-errorRuleRecursion tk k = errorStructured -- may involve both rules and oracle, so report as a rule
+errorRuleRecursion :: [String] -> Maybe TypeRep -> Maybe String -> IO a
+-- may involve both rules and oracle, so report as only rules
+errorRuleRecursion stack tk k = throwIO $ wrap $ toException $ ErrorCall $ errorStructuredContents
     "Build system error - recursion detected"
     [("Key type",fmap show tk)
     ,("Key value",k)]
     "Rules may not be recursive"
+    where
+        wrap = if null stack then id else toException . ShakeException (last stack) stack
+
+errorComplexRecursion :: [String] -> IO a
+errorComplexRecursion ks = errorStructured
+    "Build system error - indirect recursion detected"
+    [("Key value " ++ show i, Just k) | (i, k) <- zip [1..] ks]
+    "Rules may not be recursive"
 
 errorDuplicateOracle :: TypeRep -> Maybe String -> [TypeRep] -> IO a
 errorDuplicateOracle tk k tvs = errorStructured
@@ -117,10 +129,10 @@ specialIsOracleKey t = con == "OracleQ"
 -- | Error representing all expected exceptions thrown by Shake.
 --   Problems when executing rules will be raising using this exception type.
 data ShakeException = ShakeException
-        {shakeExceptionTarget :: String -- ^ The target that was being built when the exception occured.
-        ,shakeExceptionStack :: [String]  -- ^ The stack of targets, where the 'shakeExceptionTarget' is last.
-        ,shakeExceptionInner :: SomeException -- ^ The underlying exception that was raised.
-        }
+    {shakeExceptionTarget :: String -- ^ The target that was being built when the exception occured.
+    ,shakeExceptionStack :: [String]  -- ^ The stack of targets, where the 'shakeExceptionTarget' is last.
+    ,shakeExceptionInner :: SomeException -- ^ The underlying exception that was raised.
+    }
     deriving Typeable
 
 instance Exception ShakeException
