@@ -12,7 +12,8 @@ import System.Directory as IO
 import qualified System.IO.Extra as IO
 
 
-main = shaken2 test $ \args obj -> do
+main = shakenCwd test $ \args obj -> do
+    want args
     obj "norule" %> \_ ->
         need [obj "norule_isavailable"]
 
@@ -22,8 +23,11 @@ main = shaken2 test $ \args obj -> do
     [obj "failcreates", obj "failcreates2"] &%> \_ ->
         writeFile' (obj "failcreates") ""
 
-    obj "recursive" %> \out ->
-        need [out]
+    obj "recursive_" %> \out -> need [obj "intermediate_"]
+    obj "intermediate_" %> \out -> need [obj "recursive_"]
+
+    "rec1" %> \out -> need ["rec2"]
+    "rec2" %> \out -> need ["rec1"]
 
     obj "systemcmd" %> \_ ->
         cmd "random_missing_command"
@@ -49,7 +53,7 @@ main = shaken2 test $ \args obj -> do
     catcher "exception2" $ actionOnException $ return ()
 
     res <- newResource "resource_name" 1
-    obj "resource" %> \out -> do
+    obj "resource" %> \out ->
         withResource res 1 $
             need ["resource-dep"]
 
@@ -89,6 +93,16 @@ main = shaken2 test $ \args obj -> do
 
     when ("die" `elem` args) $ action $ error "death error"
 
+    obj "fresh_dir" %> \out -> liftIO $ createDirectoryIfMissing True out
+    obj "need_dir" %> \out -> do
+        liftIO $ createDirectoryIfMissing True $ obj "existing_dir"
+        need [obj "existing_dir"]
+        writeFile' out ""
+
+
+    -- not tested by default since only causes an error when idle GC is turned on
+    phony "block" $
+        liftIO $ putStrLn $ let x = x in x
 
 test build obj = do
     let crash args parts = assertException parts (build $ "--quiet" : args)
@@ -96,58 +110,59 @@ test build obj = do
     build ["--sleep"]
 
     writeFile (obj "chain.1") "x"
-    build ["$chain.3","--sleep"]
+    build ["chain.3","--sleep"]
     writeFile (obj "chain.1") "err"
-    crash ["$chain.3"] ["err_chain"]
+    crash ["chain.3"] ["err_chain"]
 
-    crash ["$norule"] ["norule_isavailable"]
-    crash ["$failcreate"] ["failcreate"]
-    crash ["$failcreates"] ["failcreates"]
-    crash ["$recursive"] ["recursive"]
-    crash ["$systemcmd"] ["systemcmd","random_missing_command"]
-    crash ["$stack1"] ["stack1","stack2","stack3","crash"]
+    crash ["norule"] ["norule_isavailable"]
+    crash ["failcreate"] ["failcreate"]
+    crash ["failcreates"] ["failcreates"]
+    crash ["recursive_"] ["recursive_","intermediate_","recursive"]
+    crash ["rec1","rec2"] ["rec1","rec2","recursive"]
+    crash ["systemcmd"] ["systemcmd","random_missing_command"]
+    crash ["stack1"] ["stack1","stack2","stack3","crash"]
 
     b <- IO.doesFileExist $ obj "staunch1"
     when b $ removeFile $ obj "staunch1"
-    crash ["$staunch1","$staunch2","-j2"] ["crash"]
+    crash ["staunch1","staunch2","-j2"] ["crash"]
     b <- IO.doesFileExist $ obj "staunch1"
     assert (not b) "File should not exist, should have crashed first"
-    crash ["$staunch1","$staunch2","-j2","--keep-going","--silent"] ["crash"]
+    crash ["staunch1","staunch2","-j2","--keep-going","--silent"] ["crash"]
     b <- IO.doesFileExist $ obj "staunch1"
     assert b "File should exist, staunch should have let it be created"
 
-    crash ["$finally1"] ["die"]
+    crash ["finally1"] ["die"]
     assertContents (obj "finally1") "1"
-    build ["$finally2"]
+    build ["finally2"]
     assertContents (obj "finally2") "1"
-    crash ["$exception1"] ["die"]
+    crash ["exception1"] ["die"]
     assertContents (obj "exception1") "1"
-    build ["$exception2"]
+    build ["exception2"]
     assertContents (obj "exception2") "0"
 
     forM_ ["finally3","finally4"] $ \name -> do
-        t <- forkIO $ ignore $ build ['$':name,"--exception"]
+        t <- forkIO $ ignore $ build [name,"--exception"]
         retry 10 $ sleep 0.1 >> assertContents (obj name) "0"
         throwTo t (IndexOutOfBounds "test")
         retry 10 $ sleep 0.1 >> assertContents (obj name) "1"
 
-    crash ["$resource"] ["cannot currently call apply","withResource","resource_name"]
+    crash ["resource"] ["cannot currently call apply","withResource","resource_name"]
 
-    build ["$overlap.foo"]
+    build ["overlap.foo"]
     assertContents (obj "overlap.foo") "overlap.*"
-    build ["$overlap.txt"]
+    build ["overlap.txt"]
     assertContents (obj "overlap.txt") "overlap.txt"
-    crash ["$overlap.txx"] ["key matches multiple rules","overlap.txx"]
-    build ["$alternative.foo","$alternative.txt"]
+    crash ["overlap.txx"] ["key matches multiple rules","overlap.txx"]
+    build ["alternative.foo","alternative.txt"]
     assertContents (obj "alternative.foo") "alternative.*"
     assertContents (obj "alternative.txt") "alternative.txt"
 
-    crash ["$tempfile"] ["tempfile-died"]
+    crash ["tempfile"] ["tempfile-died"]
     src <- readFile $ obj "tempfile"
     assertMissing src
-    build ["$tempdir"]
+    build ["tempdir"]
 
-    crash ["!die"] ["Shake","action","death error"]
+    crash ["die"] ["Shake","action","death error"]
 
     putStrLn "## BUILD errors"
     (out,_) <- IO.captureOutput $ build []
@@ -156,3 +171,6 @@ test build obj = do
     putStrLn "## BUILD errors fail1 fail2 -k -j2"
     (out,_) <- IO.captureOutput $ try_ $ build ["fail1","fail2","-k","-j2",""]
     assert ("die1" `isInfixOf` out && "die2" `isInfixOf` out) $ "Expected 'die1' and 'die2', but got: " ++ out
+
+    crash ["fresh_dir"] ["expected a file, got a directory","fresh_dir"]
+    crash ["need_dir"] ["expected a file, got a directory","existing_dir"]
