@@ -2,58 +2,69 @@
 module Test.Type(sleep, module Test.Type) where
 
 import Development.Shake hiding (copyFileChanged)
+import Development.Shake.Forward
 import Development.Shake.Rule() -- ensure the module gets imported, and thus tested
 import General.String
+import General.Extra
 import Development.Shake.FileInfo
 import Development.Shake.FilePath
+import Paths_shake
 
-import Control.Applicative
 import Control.Exception.Extra hiding (assert)
 import Control.Monad.Extra
 import Data.List
 import Data.Maybe
 import qualified Data.ByteString as BS
-import System.Directory as IO
+import System.Directory.Extra as IO
 import System.Environment.Extra
 import System.Random
 import System.Console.GetOpt
 import System.IO.Extra as IO
 import System.Time.Extra
-import System.Info.Extra
 import Prelude
 
 
-shaken
+shaken, shakenCwd
     :: (([String] -> IO ()) -> (String -> String) -> IO ())
     -> ([String] -> (String -> String) -> Rules ())
     -> IO ()
     -> IO ()
-shaken test rules sleeper = do
+shaken = shakenEx False
+shakenCwd = shakenEx True
+
+shakenEx
+    :: Bool
+    -> (([String] -> IO ()) -> (String -> String) -> IO ())
+    -> ([String] -> (String -> String) -> Rules ())
+    -> IO ()
+    -> IO ()
+shakenEx changeDir test rules sleeper = do
+    -- my debug getDataFileName (in Paths) uses a cache of the Cwd
+    -- make sure we force the cache before changing directory
+    getDataFileName ""
+
     name:args <- getArgs
     when ("--sleep" `elem` args) sleeper
     putStrLn $ "## BUILD " ++ unwords (name:args)
-    args <- return $ delete "--sleep" args
+    let forward = "--forward" `elem` args
+    args <- return $ args \\ ["--sleep","--forward"]
+    cwd <- getCurrentDirectory
     let out = "output/" ++ name ++ "/"
-    let obj x = if "/" `isPrefixOf` x then init out ++ x else out ++ x
+    let obj x | changeDir = if null x then "." else x
+              | otherwise = if "/" `isPrefixOf` x || null x then init out ++ x else out ++ x
+    let change = if changeDir then withCurrentDirectory out else id
+    let unchange act = do
+            new <- getCurrentDirectory
+            withCurrentDirectory cwd $ do act; createDirectoryIfMissing True new -- to deal with clean
     createDirectoryIfMissing True out
     case args of
         "test":extra -> do
             putStrLn $ "## TESTING " ++ name
             -- if the extra arguments are not --quiet/--loud it's probably going to go wrong
-            test (\args -> withArgs (name:args ++ extra) $ shaken test rules sleeper) obj
+            change $ test (\args -> withArgs (name:args ++ extra) $ unchange $ shakenEx changeDir test rules sleeper) obj
             putStrLn $ "## FINISHED TESTING " ++ name
+
         "clean":_ -> removeDirectoryRecursive out
-{-
-        "lint":args -> do
-            let dbfile = out ++ ".database"
-                tempfile = "output/" ++ name ++ ".database"
-            b <- IO.doesFileExist dbfile
-            when b $ renameFile dbfile tempfile
-            removeDirectoryRecursive out
-            createDirectoryIfMissing True out
-            when b $ renameFile tempfile dbfile
-            shake shakeOptions{shakeFiles=out, shakeLint=True} $ rules args (out++)
--}
 
         "perturb":args -> forever $ do
             del <- removeFilesRandom out
@@ -62,35 +73,34 @@ shaken test rules sleeper = do
             shake shakeOptions{shakeFiles=out, shakeThreads=threads, shakeVerbosity=Quiet} $ rules args (out++)
 
         args -> do
+            t <- tracker
             let (_,files,_) = getOpt Permute [] args
-            tracker <- hasTracker
+            opts <- return $ shakeOptions
+                {shakeFiles = obj ""
+                ,shakeReport = [obj "report.html"]}
+            opts <- return $ if forward then forwardOptions opts else opts
+                {shakeLint = Just t
+                ,shakeLintInside = [cwd]
+                ,shakeLintIgnore = map (cwd </>) [".cabal-sandbox//",".stack-work//"]}
             withArgs (args \\ files) $
-                shakeWithClean
-                    (removeDirectoryRecursive out)
-                    (shakeOptions{shakeFiles = out
-                                 ,shakeReport = ["output/" ++ name ++ "/report.html"]
-                                 ,shakeLint = Just $ if tracker then LintTracker else LintBasic
-                                 })
+                change $ shakeWithClean
+                    (unchange $ removeDirectoryRecursive out)
+                    opts
                     -- if you have passed sleep, supress the "no errors" warning
                     (do rules files obj; when ("--sleep" `elem` args) $ action $ return ())
 
 
-shaken2
-    :: (([String] -> IO ()) -> (String -> String) -> IO ())
-    -> ([String] -> (String -> String) -> Rules ())
-    -> IO ()
-    -> IO ()
-shaken2 test rules sleeper = shaken test rules2 sleeper
-    where
-        rules2 args obj = do
-            (objd,args) <- return $ partition ("$" `isPrefixOf`) args
-            (spec,phon) <- return $ partition ("!" `isPrefixOf`) args
-            want $ phon ++ map (obj . tail) objd
-            rules (map tail spec) obj
-
+tracker :: IO Lint
+tracker = do
+  fsatrace <- findExecutable $ "fsatrace" <.> exe
+  return $ if isJust fsatrace
+           then LintFSATrace
+           else LintBasic
 
 hasTracker :: IO Bool
-hasTracker = isJust <$> if isWindows then findExecutable "tracker.exe" else lookupEnv "FSAT"
+hasTracker = do
+  t <- tracker
+  return $ t == LintFSATrace
 
 
 shakeWithClean :: IO () -> ShakeOptions -> Rules () -> IO ()
@@ -157,7 +167,7 @@ assertException parts act = do
     case res of
         Left err -> let s = show err in forM_ parts $ \p ->
             assert (p `isInfixOf` s) $ "Incorrect exception, missing part:\nGOT: " ++ s ++ "\nWANTED: " ++ p
-        Right _ -> error "Expected an exception but succeeded"
+        Right _ -> error $ "Expected an exception containing " ++ show parts ++ ", but succeeded"
 
 
 noTest :: ([String] -> IO ()) -> (String -> String) -> IO ()
@@ -184,8 +194,8 @@ sleepFileTimeCalibrate = do
             writeFile file $ show (i,j)
             t2 <- time
             return $ if t1 == t2 then Left $ j+1 else Right ()
-    putStrLn $ "Longest file modification time lag was " ++ show (ceiling (maximum mtimes * 1000)) ++ "ms"
-    return $ sleep $ min 1 $ maximum mtimes * 2
+    putStrLn $ "Longest file modification time lag was " ++ show (ceiling (maximum' mtimes * 1000)) ++ "ms"
+    return $ sleep $ min 1 $ maximum' mtimes * 2
 
 
 removeFilesRandom :: FilePath -> IO Int
@@ -218,19 +228,4 @@ copyFileChanged :: FilePath -> FilePath -> IO ()
 copyFileChanged old new = do
     good <- IO.doesFileExist new
     good <- if not good then return False else liftM2 (==) (BS.readFile old) (BS.readFile new)
-    when (not good) $ copyFile old new
-
-
-withTemporaryDirectory :: (FilePath -> IO ()) -> IO ()
-withTemporaryDirectory act = do
-    tdir <- getTemporaryDirectory
-    bracket
-        (openTempFile tdir "shake.hs")
-        (removeFile . fst)
-        $ \(file,h) -> do
-            hClose h
-            let dir = file ++ "_"
-            bracket_ (createDirectory dir) (removeDirectoryRecursive dir) (act dir)
-
-skip :: Monad m => a -> m ()
-skip x = return ()
+    unless good $ copyFile old new
